@@ -11,6 +11,7 @@ import os
 from enum import Enum
 from functools import partial, reduce
 from importlib.metadata import PackageNotFoundError, version
+from typing import List
 
 import dask
 import magicgui.widgets as w
@@ -18,6 +19,7 @@ import napari
 import numpy as np
 import pandas as pd
 import qtpy.QtWidgets as q
+import tifffile as tf
 import yaml
 from magicgui import magic_factory
 from matplotlib.backends.backend_qt5agg import (
@@ -55,6 +57,7 @@ formatter = logging.Formatter(
 ff = logging.FileHandler("napari-segment.log")
 ff.setFormatter(formatter)
 logger.addHandler(ff)
+logger.setLevel(logging.WARNING)
 
 
 class SegmentStack(q.QWidget):
@@ -73,7 +76,7 @@ class SegmentStack(q.QWidget):
         self.viewer = napari_viewer
 
         self.input = w.ComboBox(
-            label="BF data",
+            label="Input",
             annotation=Image,
             choices=self.update_images(),
         )
@@ -87,6 +90,7 @@ class SegmentStack(q.QWidget):
         self.btn_make_manual_labels = w.PushButton(
             text="Clone for manual correction"
         )
+
         self.btn_make_manual_labels.clicked.connect(self.make_manual_layer)
 
         self.stat_layer_selector_container = w.Container(
@@ -95,17 +99,19 @@ class SegmentStack(q.QWidget):
         )
 
         self.binning_widget = w.RadioButtons(
-            label="binning",
+            label="Binning",
             choices=[2**n for n in range(4)],
             value=4,
             orientation="horizontal",
         )
 
-        self.thr = w.FloatSlider(
-            label="Threshold", min=0.1, max=0.9, value=0.4
+        self.threshold_widget = w.FloatSlider(
+            label="Threshold", min=0.0, max=1.0, value=0.4
         )
 
-        self.erode = w.SpinBox(label="erode", min=0, max=10, value=0)
+        self.erode_widget = w.SpinBox(
+            label="Erode (px)", min=0, max=10, value=0
+        )
 
         self.use = w.RadioButtons(
             label="Use",
@@ -115,7 +121,9 @@ class SegmentStack(q.QWidget):
             allow_multiple=True,
         )
 
-        self.smooth = w.SpinBox(label="smooth", min=0, max=10, value=2)
+        self.smooth_widget = w.SpinBox(
+            label="Smooth sigma (px)", min=0, max=10, value=2
+        )
 
         self.min_diam = w.Slider(
             label="Min_diameter",
@@ -132,7 +140,7 @@ class SegmentStack(q.QWidget):
         )
 
         self.btn_save_params = q.QPushButton("Save params!")
-        self.btn_save_csv = q.QPushButton("Save csv!")
+        self.btn_save_csv = q.QPushButton("Save csv, params, labels!")
 
         self.check_auto_plot = w.CheckBox(label="Auto Update")
         self.check_auto_plot.changed.connect(self.plot_stats)
@@ -152,7 +160,7 @@ class SegmentStack(q.QWidget):
             value=self.pixel_size,
         )  # bind=self.set_pixel_size)
         self.pixel_unit_widget = w.LineEdit(
-            label="unit", value=self.pixel_unit
+            label="Unit", value=self.pixel_unit
         )  # , bind=self.set_pixel_unit)
         self.pixel_block = w.Container(
             widgets=[self.pixel_size_widget, self.pixel_unit_widget],
@@ -177,26 +185,27 @@ class SegmentStack(q.QWidget):
             range(10), [0] * 10, "o", picker=True, pickradius=5
         )
 
-        self.container = w.Container(
-            label="Container",
-            widgets=[
-                self.input,
-                w.Label(label="Prepocessing"),
-                self.binning_widget,
-                self.use,
-                w.Label(label="Detection"),
-                self.smooth,
-                self.thr,
-                self.erode,
-                w.Label(label="Filters"),
-                self.min_diam,
-                self.max_diam,
-                self.max_ecc,
-            ],
+        self.setLayout(q.QVBoxLayout())
+
+        pre_block = make_group(
+            title="Preprocessing",
+            widgets=[self.binning_widget, self.use, self.smooth_widget],
         )
 
-        self.setLayout(q.QVBoxLayout())
-        self.layout().addWidget(self.container.native)
+        det_block = make_group(
+            title="Thresholding",
+            widgets=[self.threshold_widget, self.erode_widget],
+        )
+
+        fil_block = make_group(
+            title="Filtering",
+            widgets=[self.min_diam, self.max_diam, self.max_ecc],
+        )
+
+        self.layout().addWidget(w.Container(widgets=[self.input]).native)
+        self.layout().addWidget(pre_block)
+        self.layout().addWidget(det_block)
+        self.layout().addWidget(fil_block)
         self.layout().addWidget(self.btn_save_params)
         self.layout().addWidget(self.update_plot_container.native)
         self.layout().addWidget(self.stat_layer_selector_container.native)
@@ -235,10 +244,20 @@ class SegmentStack(q.QWidget):
             if isinstance(layer, Image)
         ]
 
-    def make_manual_layer(self):
+    def make_manual_layer(self, data: np.ndarray = None):
         try:
-            clone = self.selected_labels.compute()
-            self.viewer.add_labels(
+            self.viewer.layers["Manual Labels"].remove()
+            logger.debug("Old Manual labels layer deleted")
+        except Exception as e:
+            logger.debug(f"No manual labels present, nothing to delete {e}")
+
+        try:
+            if isinstance(data, np.ndarray):
+                clone = data
+            else:
+                clone = self.selected_labels.compute()
+
+            self.manual_layer = self.viewer.add_labels(
                 data=clone,
                 name="Manual Labels",
                 scale=self.scale,
@@ -247,8 +266,10 @@ class SegmentStack(q.QWidget):
                     "source": self.viewer.layers["selected labels"],
                 },
             )
-            self.stat_layer_selector.choices = self.update_labels()[::-1]
-
+            self.stat_layer_selector.choices = ["Manual Labels"]
+            self.viewer.layers["selected labels"].visible = False
+            self.viewer.layers["Detections"].visible = False
+            self.viewer.layers["Preprocessing"].visible = False
             self.plot_stats(force=True)
 
         except Exception as e:
@@ -284,15 +305,22 @@ class SegmentStack(q.QWidget):
         self.move_step(xdata[ind])
 
     def _invert(self, data2D):
-        return 1 - norm01(gaussian_filter(data2D, self.smooth.value))
+        logger.debug(
+            f"Making inverted intensity with sigma {self.smooth_widget.value}"
+        )
+        return 1 - norm01(gaussian_filter(data2D, self.smooth_widget.value))
 
     def _grad(self, data2D):
-        return get_gradient(data2D, smooth=self.smooth.value)
+        logger.debug(f"Making gradient with sigma {self.smooth_widget.value}")
+        return get_gradient(data2D, smooth=self.smooth_widget.value)
 
     def _gdif(self, data2D):
-        return gaussian_filter(data2D, self.smooth.value) - gaussian_filter(
-            data2D, self.smooth.value + 2
+        logger.debug(
+            f"Making Gauss difference with sigma {self.smooth_widget.value}"
         )
+        return gaussian_filter(
+            data2D, self.smooth_widget.value
+        ) - gaussian_filter(data2D, self.smooth_widget.value + 2)
 
     def set_pixel_size(self, value):
         self.pixel_size = value
@@ -315,14 +343,8 @@ class SegmentStack(q.QWidget):
             self.path = ""
 
         self.binning = self.binning_widget.value
-        try:
-            self.data = self.viewer.layers[self.input.current_choice].data[
-                ..., :: self.binning, :: self.binning
-            ]
-            logger.debug(f"data after binning: {self.data}")
-
-        except KeyError:
-            return
+        self.raw_data = self.viewer.layers[self.input.current_choice].data
+        self.set_data()
 
         try:
             pixel_size = self.viewer.layers[
@@ -338,9 +360,7 @@ class SegmentStack(q.QWidget):
 
         logger.debug(f"pixel size: {self.pixel_size} {self.pixel_unit}")
 
-        self.scale = np.ones((len(self.data.shape),))
-        self.scale[-2:] = self.binning
-        logger.debug(f"Computed scale for napari {self.scale}")
+        self.set_scale()
 
         if isinstance(self.data, np.ndarray):
             chunksize = np.ones(len(self.data.shape))
@@ -403,17 +423,35 @@ class SegmentStack(q.QWidget):
         )
         self.threshold()
 
+    def set_data(self):
+        try:
+            self.data = self.raw_data[..., :: self.binning, :: self.binning]
+            logger.debug(f"data after binning: {self.data}")
+
+        except KeyError:
+            logger.error(
+                f"unable to bin the data {self.raw_data} \
+                    with binning {self.binning}"
+            )
+
+    def set_scale(self):
+        self.scale = np.ones((len(self.data.shape),))
+        self.scale[-2:] = self.binning
+        logger.debug(f"Computed scale for napari {self.scale}")
+
     def threshold(self):
         logger.debug("Start thresholding step")
         if not self.input.current_choice:
             return
         logger.debug(
-            f"Thresholding with the thr={self.thr.value} \
-            and erode={self.erode.value}"
+            f"Thresholding with the thr={self.threshold_widget.value} \
+            and erode={self.erode_widget.value}"
         )
         self.labels = self.smooth_gradient.map_blocks(
             partial(
-                threshold_gradient, thr=self.thr.value, erode=self.erode.value
+                threshold_gradient,
+                thr=self.threshold_widget.value,
+                erode=self.erode_widget.value,
             ),
             dtype=np.int32,
         )
@@ -561,18 +599,25 @@ class SegmentStack(q.QWidget):
             logger.error(f"Error saving csv: {e}")
             show_error(f"Error saving csv: {e}")
 
-    def save_params(self):
+        if self.stat_layer_selector.value == "Manual Labels":
+            out = self.manual_layer.save(self.path + ".labels.tif")
+            show_info(f"Manual labels saved {out[0]}")
+            logger.info(f"Manual labels saved {out[0]}")
+            self.save_params(manual_labels=os.path.basename(out[0]))
+
+    def save_params(self, **kwargs):
         data = {
             "binning": self.binning,
             "use": self.use.value,
-            "smooth": self.smooth.value,
-            "thr": self.thr.value,
-            "erode": self.erode.value,
+            "smooth": self.smooth_widget.value,
+            "thr": self.threshold_widget.value,
+            "erode": self.erode_widget.value,
             "min_diameter": self.min_diam.value,
             "max_diameter": self.max_diam.value,
             "max_ecc": self.max_ecc.value,
             "pixel_size": self.pixel_size,
             "pixel_unit": self.pixel_unit,
+            **kwargs,
         }
         try:
 
@@ -587,6 +632,7 @@ class SegmentStack(q.QWidget):
         except Exception as e:
             show_error(f"Saving parameters into {new_name} failed: {e}")
             logger.error(f"Saving parameters into {new_name} failed: {e}")
+
         with open((ff := ".latest.params.yaml"), "w") as f:
             yaml.safe_dump(data, f)
             logger.info(f"Parameters saves into {ff}")
@@ -625,23 +671,36 @@ class SegmentStack(q.QWidget):
         try:
             self.binning_widget.value = data["binning"]
             self.use.value = data["use"]
-            self.smooth.value = data["smooth"]
-            self.thr.value = data["thr"]
-            self.erode.value = data["erode"]
+            self.smooth_widget.value = data["smooth"]
+            self.threshold_widget.value = data["thr"]
+            self.erode_widget.value = data["erode"]
             self.min_diam.value = data["min_diameter"]
             self.max_diam.value = data["max_diameter"]
             self.max_ecc.value = data["max_ecc"]
             self.pixel_size = data["pixel_size"]
             self.pixel_unit = data["pixel_unit"]
+
         except Exception as e:
             show_error(f"Restore settings failed, {e}")
             logger.error(f"Restore settings failed, {e}")
 
+        self.preprocess()
+
+        try:
+            manual_labels_path = os.path.join(
+                os.path.dirname(self.path), data["manual_labels"]
+            )
+            manual_labels = tf.imread(manual_labels_path)
+            self.make_manual_layer(data=manual_labels)
+        except Exception as e:
+            show_error(f"Unable to restore manual labels {e}")
+            logger.error(f"Unable to restore manual labels {e}")
+
         self.binning_widget.changed.connect(self.preprocess)
-        self.thr.changed.connect(self.threshold)
-        self.erode.changed.connect(self.threshold)
+        self.threshold_widget.changed.connect(self.threshold)
+        self.erode_widget.changed.connect(self.threshold)
         self.use.changed.connect(self.preprocess)
-        self.smooth.changed.connect(self.preprocess)
+        self.smooth_widget.changed.connect(self.preprocess)
         self.min_diam.changed.connect(self.update_out)
         self.max_diam.changed.connect(self.update_out)
         self.max_ecc.changed.connect(self.update_out)
@@ -665,6 +724,19 @@ class SegmentStack(q.QWidget):
             )
 
 
+def make_group(
+    title: str = "Group",
+    widgets: List[w.Widget] = [],
+    layout: str = "vertical",
+):
+    group = q.QGroupBox(title=title)
+    layout = q.QVBoxLayout()
+    cnt = w.Container(widgets=widgets, layout=layout)
+    layout.addWidget(cnt.native)
+    group.setLayout(layout)
+    return group
+
+
 def norm01(data):
     d = data.copy()
     if d.max() == 0:
